@@ -1,4 +1,4 @@
-# Copyright 2014 Microsoft Corporation
+# Copyright 2018 Microsoft Corporation
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -12,62 +12,91 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 #
-# Requires Python 2.4+ and Openssl 1.0+
+# Requires Python 2.6+ and Openssl 1.0+
 #
 
+import atexit
+import datetime
+import json
 import os
 import sys
-import traceback
-import atexit
-import json
 import time
-import datetime
-import threading
-import platform
+import traceback
 
-from datetime import datetime, timedelta
+from datetime import datetime
 
 import azurelinuxagent.common.conf as conf
 import azurelinuxagent.common.logger as logger
 
-from azurelinuxagent.common.exception import EventError, ProtocolError
+from azurelinuxagent.common.exception import EventError
 from azurelinuxagent.common.future import ustr
 from azurelinuxagent.common.protocol.restapi import TelemetryEventParam, \
-    TelemetryEventList, \
     TelemetryEvent, \
-    set_properties, get_properties
-from azurelinuxagent.common.version import DISTRO_NAME, DISTRO_VERSION, \
-    DISTRO_CODE_NAME, AGENT_VERSION, \
-    CURRENT_AGENT, CURRENT_VERSION
+    get_properties
+from azurelinuxagent.common.utils import fileutil, textutil
+from azurelinuxagent.common.version import CURRENT_VERSION
 
 _EVENT_MSG = "Event: name={0}, op={1}, message={2}, duration={3}"
 
+
 class WALAEventOperation:
     ActivateResourceDisk = "ActivateResourceDisk"
+    AgentBlacklisted = "AgentBlacklisted"
+    AgentEnabled = "AgentEnabled"
+    ArtifactsProfileBlob = "ArtifactsProfileBlob"
     AutoUpdate = "AutoUpdate"
+    CustomData = "CustomData"
+    CGroupsLimitsCrossed = "CGroupsLimitsCrossed"
+    Deploy = "Deploy"
     Disable = "Disable"
+    Downgrade = "Downgrade"
     Download = "Download"
     Enable = "Enable"
+    ExtensionProcessing = "ExtensionProcessing"
     Firewall = "Firewall"
+    GetArtifactExtended = "GetArtifactExtended"
     HealthCheck = "HealthCheck"
+    HealthObservation = "HealthObservation"
     HeartBeat = "HeartBeat"
     HostPlugin = "HostPlugin"
+    HostPluginHeartbeat = "HostPluginHeartbeat"
+    HostPluginHeartbeatExtended = "HostPluginHeartbeatExtended"
+    HttpErrors = "HttpErrors"
+    ImdsHeartbeat = "ImdsHeartbeat"
     Install = "Install"
+    InitializeCGroups = "InitializeCGroups"
     InitializeHostPlugin = "InitializeHostPlugin"
+    Log = "Log"
+    Partition = "Partition"
     ProcessGoalState = "ProcessGoalState"
     Provision = "Provision"
+    ProvisionGuestAgent = "ProvisionGuestAgent"
+    RemoteAccessHandling = "RemoteAccessHandling"
     ReportStatus = "ReportStatus"
+    ReportStatusExtended = "ReportStatusExtended"
     Restart = "Restart"
+    SequenceNumberMismatch = "SequenceNumberMismatch"
+    SetCGroupsLimits = "SetCGroupsLimits"
+    SkipUpdate = "SkipUpdate"
     UnhandledError = "UnhandledError"
     UnInstall = "UnInstall"
+    Unknown = "Unknown"
     Upgrade = "Upgrade"
     Update = "Update"
 
 
+SHOULD_ENCODE_MESSAGE_LEN = 80
+SHOULD_ENCODE_MESSAGE_OP = [
+    WALAEventOperation.Disable,
+    WALAEventOperation.Enable,
+    WALAEventOperation.Install,
+    WALAEventOperation.UnInstall,
+]
+
 class EventStatus(object):
     EVENT_STATUS_FILE = "event_status.json"
 
-    def __init__(self, status_dir=conf.get_lib_dir()):
+    def __init__(self):
         self._path = None
         self._status = {}
 
@@ -82,7 +111,7 @@ class EventStatus(object):
         event = self._event_name(name, version, op)
         if event not in self._status:
             return True
-        return self._status[event] == True
+        return self._status[event] is True
 
     def initialize(self, status_dir=conf.get_lib_dir()):
         self._path = os.path.join(status_dir, EventStatus.EVENT_STATUS_FILE)
@@ -90,7 +119,7 @@ class EventStatus(object):
 
     def mark_event_status(self, name, version, op, status):
         event = self._event_name(name, version, op)
-        self._status[event] = (status == True)
+        self._status[event] = (status is True)
         self._save()
 
     def _event_name(self, name, version, op):
@@ -113,6 +142,7 @@ class EventStatus(object):
         except Exception as e:
             logger.warn("Exception occurred saving event status: {0}".format(e))
 
+
 __event_status__ = EventStatus()
 __event_status_operations__ = [
         WALAEventOperation.AutoUpdate,
@@ -120,9 +150,45 @@ __event_status_operations__ = [
     ]
 
 
+def _encode_message(op, message):
+    """
+    Gzip and base64 encode a message based on the operation.
+
+    The intent of this message is to make the logs human readable and include the
+    stdout/stderr from extension operations.  Extension operations tend to generate
+    a lot of noise, which makes it difficult to parse the line-oriented waagent.log.
+    The compromise is to encode the stdout/stderr so we preserve the data and do
+    not destroy the line oriented nature.
+
+    The data can be recovered using the following command:
+
+      $ echo '<encoded data>' | base64 -d | pigz -zd
+
+    You may need to install the pigz command.
+
+    :param op: Operation, e.g. Enable or Install
+    :param message: Message to encode
+    :return: gzip'ed and base64 encoded message, or the original message
+    """
+
+    if len(message) == 0:
+        return message
+
+    if op not in SHOULD_ENCODE_MESSAGE_OP:
+        return message
+
+    try:
+        return textutil.compress(message)
+    except Exception:
+        # If the message could not be encoded a dummy message ('<>') is returned.
+        # The original message was still sent via telemetry, so all is not lost.
+        return "<>"
+
+
 def _log_event(name, op, message, duration, is_success=True):
     global _EVENT_MSG
 
+    message = _encode_message(op, message)
     if not is_success:
         logger.error(_EVENT_MSG, name, op, message, duration)
     else:
@@ -139,9 +205,7 @@ class EventLogger(object):
             logger.warn("Cannot save event -- Event reporter is not initialized.")
             return
 
-        if not os.path.exists(self.event_dir):
-            os.mkdir(self.event_dir)
-            os.chmod(self.event_dir, 0o700)
+        fileutil.mkdir(self.event_dir, mode=0o700)
 
         existing_events = os.listdir(self.event_dir)
         if len(existing_events) >= 1000:
@@ -171,7 +235,7 @@ class EventLogger(object):
             (self.periodic_events[h] + delta) <= datetime.now()
 
     def add_periodic(self,
-        delta, name, op="", is_success=True, duration=0,
+        delta, name, op=WALAEventOperation.Unknown, is_success=True, duration=0,
         version=CURRENT_VERSION, message="", evt_type="",
         is_internal=False, log_event=True, force=False):
 
@@ -186,7 +250,7 @@ class EventLogger(object):
 
     def add_event(self,
                   name,
-                  op="",
+                  op=WALAEventOperation.Unknown,
                   is_success=True,
                   duration=0,
                   version=CURRENT_VERSION,
@@ -195,10 +259,14 @@ class EventLogger(object):
                   is_internal=False,
                   log_event=True):
 
-        if not is_success or log_event:
+        if (not is_success) and log_event:
             _log_event(name, op, message, duration, is_success=is_success)
 
-        event = TelemetryEvent(1, "69B669B9-4AF8-4C50-BDC4-6006FA76E975")
+        self._add_event(duration, evt_type, is_internal, is_success, message, name, op, version, eventId=1)
+        self._add_event(duration, evt_type, is_internal, is_success, message, name, op, version, eventId=6)
+
+    def _add_event(self, duration, evt_type, is_internal, is_success, message, name, op, version, eventId):
+        event = TelemetryEvent(eventId, "69B669B9-4AF8-4C50-BDC4-6006FA76E975")
         event.parameters.append(TelemetryEventParam('Name', name))
         event.parameters.append(TelemetryEventParam('Version', str(version)))
         event.parameters.append(TelemetryEventParam('IsInternal', is_internal))
@@ -215,22 +283,82 @@ class EventLogger(object):
         except EventError as e:
             logger.error("{0}", e)
 
+    def add_log_event(self, level, message):
+        # By the time the message has gotten to this point it is formatted as
+        #
+        #   YYYY/MM/DD HH:mm:ss.fffffff LEVEL <text>.
+        #
+        # The timestamp and the level are redundant, and should be stripped.
+        # The logging library does not schematize this data, so I am forced
+        # to parse the message.  The format is regular, so the burden is low.
+
+        parts = message.split(' ', 3)
+        msg = parts[3] if len(parts) == 4 \
+            else message
+
+        event = TelemetryEvent(7, "FFF0196F-EE4C-4EAF-9AA5-776F622DEB4F")
+        event.parameters.append(TelemetryEventParam('EventName', WALAEventOperation.Log))
+        event.parameters.append(TelemetryEventParam('CapabilityUsed', logger.LogLevel.STRINGS[level]))
+        event.parameters.append(TelemetryEventParam('Context1', msg))
+        event.parameters.append(TelemetryEventParam('Context2', ''))
+        event.parameters.append(TelemetryEventParam('Context3', ''))
+
+        data = get_properties(event)
+        try:
+            self.save_event(json.dumps(data))
+        except EventError:
+            pass
+
+    def add_metric(self, category, counter, instance, value, log_event=False):
+        """
+        Create and save an event which contains a telemetry event.
+
+        :param str category: The category of metric (e.g. "cpu", "memory")
+        :param str counter: The specific metric within the category (e.g. "%idle")
+        :param str instance: For instanced metrics, the instance identifier (filesystem name, cpu core#, etc.)
+        :param value: Value of the metric
+        :param bool log_event: If true, log the collected metric in the agent log
+        """
+        if log_event:
+            from azurelinuxagent.common.version import AGENT_NAME
+            message = "Metric {0}/{1} [{2}] = {3}".format(category, counter, instance, value)
+            _log_event(AGENT_NAME, "METRIC", message, 0)
+
+        event = TelemetryEvent(4, "69B669B9-4AF8-4C50-BDC4-6006FA76E975")
+        event.parameters.append(TelemetryEventParam('Category', category))
+        event.parameters.append(TelemetryEventParam('Counter', counter))
+        event.parameters.append(TelemetryEventParam('Instance', instance))
+        event.parameters.append(TelemetryEventParam('Value', value))
+
+        data = get_properties(event)
+        try:
+            self.save_event(json.dumps(data))
+        except EventError as e:
+            logger.error("{0}", e)
+
 
 __event_logger__ = EventLogger()
 
 
 def elapsed_milliseconds(utc_start):
-    d = datetime.utcnow() - utc_start
+    now = datetime.utcnow()
+    if now < utc_start:
+        return 0
+
+    d = now - utc_start
     return int(((d.days * 24 * 60 * 60 + d.seconds) * 1000) + \
                     (d.microseconds / 1000.0))
 
-def report_event(op, is_success=True, message=''):
+
+def report_event(op, is_success=True, message='', log_event=True):
     from azurelinuxagent.common.version import AGENT_NAME, CURRENT_VERSION
     add_event(AGENT_NAME,
               version=CURRENT_VERSION,
               is_success=is_success,
               message=message,
-              op=op)
+              op=op,
+              log_event=log_event)
+
 
 def report_periodic(delta, op, is_success=True, message=''):
     from azurelinuxagent.common.version import AGENT_NAME, CURRENT_VERSION
@@ -240,7 +368,28 @@ def report_periodic(delta, op, is_success=True, message=''):
               message=message,
               op=op)
 
-def add_event(name, op="", is_success=True, duration=0, version=CURRENT_VERSION,
+
+def report_metric(category, counter, instance, value, log_event=False, reporter=__event_logger__):
+    """
+    Send a telemetry event reporting a single instance of a performance counter.
+    :param str category: The category of the metric (cpu, memory, etc)
+    :param str counter: The name of the metric ("%idle", etc)
+    :param str instance: For instanced metrics, the identifier of the instance. E.g. a disk drive name, a cpu core#
+    :param     value: The value of the metric
+    :param bool log_event: If True, log the metric in the agent log as well
+    :param EventLogger reporter: The EventLogger instance to which metric events should be sent
+    """
+    if reporter.event_dir is None:
+        from azurelinuxagent.common.version import AGENT_NAME
+        logger.warn("Cannot report metric event -- Event reporter is not initialized.")
+        message = "Metric {0}/{1} [{2}] = {3}".format(category, counter, instance, value)
+        _log_event(AGENT_NAME, "METRIC", message, 0)
+        return
+    reporter.add_metric(category, counter, instance, value, log_event)
+
+
+def add_event(name, op=WALAEventOperation.Unknown, is_success=True, duration=0,
+              version=CURRENT_VERSION,
               message="", evt_type="", is_internal=False, log_event=True,
               reporter=__event_logger__):
     if reporter.event_dir is None:
@@ -255,8 +404,17 @@ def add_event(name, op="", is_success=True, duration=0, version=CURRENT_VERSION,
             version=str(version), message=message, evt_type=evt_type,
             is_internal=is_internal, log_event=log_event)
 
+
+def add_log_event(level, message, reporter=__event_logger__):
+    if reporter.event_dir is None:
+        return
+
+    reporter.add_log_event(level, message)
+
+
 def add_periodic(
-    delta, name, op="", is_success=True, duration=0, version=CURRENT_VERSION,
+    delta, name, op=WALAEventOperation.Unknown, is_success=True, duration=0,
+    version=CURRENT_VERSION,
     message="", evt_type="", is_internal=False, log_event=True, force=False,
     reporter=__event_logger__):
     if reporter.event_dir is None:
@@ -269,10 +427,12 @@ def add_periodic(
         version=str(version), message=message, evt_type=evt_type,
         is_internal=is_internal, log_event=log_event, force=force)
 
+
 def mark_event_status(name, version, op, status):
     if op in __event_status_operations__:
         __event_status__.mark_event_status(name, version, op, status)
 
+
 def should_emit_event(name, version, op, status):
     return \
         op not in __event_status_operations__ or \
@@ -280,12 +440,15 @@ def should_emit_event(name, version, op, status):
         not __event_status__.event_marked(name, version, op) or \
         __event_status__.event_succeeded(name, version, op) != status
 
+
 def init_event_logger(event_dir):
     __event_logger__.event_dir = event_dir
 
+
 def init_event_status(status_dir):
     __event_status__.initialize(status_dir)
 
+
 def dump_unhandled_err(name):
     if hasattr(sys, 'last_type') and hasattr(sys, 'last_value') and \
             hasattr(sys, 'last_traceback'):
